package dareka;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.ConnectException;
import java.net.Socket;
import java.net.SocketException;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.ClosedSelectorException;
import java.nio.channels.SocketChannel;
import java.nio.channels.UnresolvedAddressException;
import java.util.ArrayList;
import java.util.List;
import java.util.Observable;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.CloseUtil;
import dareka.common.Config;
import dareka.common.HttpIOException;
import dareka.common.Logger;
import dareka.extensions.RequestFilter;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.Processor;
import dareka.processor.Resource;
import dareka.processor.StringResource;

public class ConnectionManager extends Observable implements Runnable {
    private SocketWrapper browser; // [nl]
    private Config config;
    private String processingURI;
    private volatile Resource processingResource;
    private volatile boolean stopped = false;

    private List<ProcessorEntry> processorEntries =
            new ArrayList<ProcessorEntry>();
    //[nl] NGXgtB^
    private List<RequestFilter> filterEntries =
            new ArrayList<RequestFilter>();

    public ConnectionManager(Config config, Socket browser) {
        if (config == null) {
            throw new IllegalArgumentException("config must not be null");
        }
        if (browser == null) {
            throw new IllegalArgumentException("browser must not be null");
        }

        this.config = config;
        this.browser = new SocketWrapper(browser); // [nl]
    }

    public void run() {
        try {
            while (processAPairOfMessages()) {
                // loop until the method returns false.
                processingURI = null;
            }
            Logger.debugWithThread("loop end");
        } catch (ConnectException e) {
            // AEgoEhɐڑs
            Logger.debugWithThread(e);
            printWarning(e);
        } catch (UnresolvedAddressException e) {
            // ݂ȂhChttpsڑ悤Ƃ
            Logger.debugWithThread(e);
            printWarning(e);
        } catch (SocketException e) {
            Logger.debugWithThread(e);

            // Connection reset ͂悭̂Œʏ̓OɏoȂ
            if (!isConnectionReset(e)) {
                printWarning(e);
            }
        } catch (HttpIOException e) {
            Logger.debugWithThread(e.toString());     // [nl] pɂɏôŒZk
        } catch (IOException e) {
            // NIOgĂConnection resetł͂Ȃȉ̃bZ[W
            // IOExceptionɂȂB
            // u̐ڑ̓[g zXgɋIɐؒf܂Bv
            // umꂽڑzXg Rs[^̃\EgEFAɂĒ~܂Bv
            // UnknownHostExceptionʏ탍Oɂ͏oȂ
            if (isConnectionReset(e) || isBrokenPipe(e)) {
                Logger.debugWithThread(e.toString()); // [nl] pɂɏôŒZk
            } else {
                Logger.debugWithThread(e);
            }
        } catch (CancelledKeyException e) {
            // read()ɕʃXbhstop()œǂݍ݂𒆒f
            // CancelledKeyExceptionɂȂB
            // Javas̎ˑB
            // Ӑ}G[Ȃ̂ŃOɂ͏oȂB
            Logger.debugWithThread(e);
        } catch (ClosedSelectorException e) {
            // ʃXbhstop()Selectorclose()
            // ClosedSelectorExceptionɂȂB
            // Ӑ}G[Ȃ̂ŃOɂ͏oȂB
            Logger.debugWithThread(e);
        } catch (Exception e) {
            Logger.debugWithThread(e);
            printWarningWithStacktrace(e);
        } finally {
            if (!browser.isClosed()) {
                consumeBrowserInput();
                CloseUtil.close(browser);
            }

            notifyCompletion();
        }
    }

    private void printWarning(Exception e) {
//        // [nl]
//        if (e.getMessage() == null || !e.getMessage().startsWith("Connection timed out")) {
//          e.printStackTrace();
//        }
        Logger.warning("failed to process: " + processingURI + "\n\t"
                + e.toString());
    }

    private void printWarningWithStacktrace(Exception e) {
        Logger.warning("failed to process: " + processingURI + "\n\t"
                + getStackTrace(e));
    }
    
    private String getStackTrace(Exception e) {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        e.printStackTrace(pw);
        pw.flush();
        return sw.toString();
    }

    private boolean isConnectionReset(/*Socket*/Exception e) { // [nl]
        // ̔@ł́c
        String message = e.getMessage();
        return message != null && (message.startsWith("Connection reset")
            // Java7ŉ̂[JCYꂽbZ[WԂĂ̂Łc
            || message.startsWith("ڑ肩烊Zbg܂")
            || message.startsWith("Software caused connection abort"));
    }

    private boolean isBrokenPipe(Exception e) { // [nl]
        String message = e.getMessage();
        return message != null && message.startsWith("Broken pipe");
    }

    /**
     * ]ȃf[^𑗂ĂuEU΍B
     */
    private void consumeBrowserInput() {
        try {
            // java.net APIŏB
            // ̂߂ɔubN[hB
            SocketChannel bc = browser.getChannel();
            // bcnullɂ͂ȂȂB
            bc.configureBlocking(true);

            // ̓ǂݍ݂ŃfbhbN邽ߏo͂͒~B
            // (FIN𑗐M)
            browser.shutdownOutput();

            // IEPOSTNGXg̎ɗ]CRLF𑗂ėĂ̂
            // IEM𐳏Ił悤ɓǂݔ΂ĂB
            // (FIN̎M܂ő҂)
            // ǂݔ΂ĂȂSocket#close()
            // ubZ[WT[o[ɐڑł܂łBvȂǂɂȂ
            // Ql: IE]CRLF𑗐M邱ƂɂĐG
            // Ă̕
            // http://support.microsoft.com/kb/823099/
            while (browser.getInputStream().read() != -1) {
                // no nothing
            }
        } catch (Exception e) {
            // IOExceptionConnection resetñG[CancelledKeyException
            // B
            // resetĂꍇ͂ł܂OɂȂ邪A
            // ۂread()Ă݂ȂƋʂtȂ̂ŎdȂB
            Logger.debugWithThread(e.toString() + "(consuming)");
        }
    }

    private boolean processAPairOfMessages() throws IOException {
        browser.deleteInputStreamBuffer(); // [nl]
        HttpRequestHeader requestHeader =
                new HttpRequestHeader(browser.getInputStream());
        processingURI = requestHeader.getURI();

        Logger.debugWithThread(requestHeader.getMethod() + " "
                + requestHeader.getURI());
        if (Boolean.getBoolean("showRequestHeader")) { // [nl]
            String header = requestHeader.toString();
            Logger.info(header.substring(0, header.length() - 2));
        }

        // [nl] ݒt@C̍XV`FbN
        // processAPairOfMessages()ƂĂ̐Uł͂Ȃ
        // new HttpRequestHeader͓T^IȃubNꏊȂ̂
        // ̒ɏƂɂB
        if (!Main.isDirectoryWatching()) {
            config.reload();
        }

        // [nl] NGXgwb_̏
        for (RequestFilter filter : filterEntries) {
            int ret = filter.onRequest(requestHeader);
            // hbvꍇ͋Iɐڑ؂
            if (ret == RequestFilter.DROP) {
                Logger.debugWithThread("request dropped");
                return false;
            }
        }

        // [nl] [J\[X̏
        Resource localResource = getLocalResource(requestHeader);
        if (localResource != null) {
            return localResource.transferTo(browser, requestHeader, config);
        }

        // [nl] Op[^̎擾Ə
        String xnlParams = requestHeader.getMessageHeader(HttpHeader.X_NICOCACHE_NL);
        if (xnlParams != null) {
            requestHeader.removeMessageHeader(HttpHeader.X_NICOCACHE_NL);
        }

        // [nl] POSTȂ̓{fBobt@Očėpł悤
        if (HttpHeader.POST.equals(requestHeader.getMethod())) {
            long contentLength = requestHeader.getContentLength();
            if (0L < contentLength && contentLength < Integer.MAX_VALUE) {
                browser.newInputStreamBuffer((int) contentLength);
            }
        }

        // ΉProcessorTď
        for (ProcessorEntry entry : processorEntries) {
            if (isMatchToEntry(entry, requestHeader)) {
                // [nl] ProcessornullԂ͎Processorŏ
                //      ExtensionőIIProcessorꍇɗL
                browser.resetInputStreamBuffer();
                Resource resource =
                    entry.getProcessor().onRequest(requestHeader, browser);
                if (resource == null) continue;
                boolean canContinue = useResource(requestHeader, resource);

                Logger.debugWithThread("end");

                // [nl] g̏ꍇ͑҂ĂʂȂ̂ŌpȂ
                if (xnlParams != null && xnlParams.contains(HttpHeader.XNL_PROXY_MYSELF))
                    canContinue = false;
                return canContinue;
            }
        }

        throw new HttpIOException("request cannot be processed:\r\n"
                + requestHeader);
    }

    // [nl] [Jւ̃NGXgȂΉResourceԂ
    private Resource getLocalResource(HttpRequestHeader requestHeader
            ) throws IOException {
        if (requestHeader.getHost().equals("LOCAL")) {
            File localFile;
            String path = requestHeader.getPath();
            if (path.length() > 1) {
                localFile = new File(path.substring(1));
            } else {
                localFile = new File("index.html");
                if (!localFile.exists()) {
                    Resource r = new StringResource(Main.VER_STRING);
                    r.setResponseHeader(HttpHeader.CONTENT_TYPE,
                            "text/plain; charset=UTF-8");
                    return r;
                }
            }
            return Main.getRewriterProcessor().localRewriter(
                    requestHeader.getURI(), localFile, requestHeader);
        }
        if (requestHeader.getURI().startsWith("http")) {
            return null; // httpNGXgȂ瑦Ԃ
        }
        String type = requestHeader.getHost().toLowerCase();
        if ("redirect".equals(type)) {
            String path = requestHeader.getPath();
            if (path.length() > 1) {
                path = path.substring(1);
            }
            return StringResource.getRedirect(path);
        } else if ("notmodified".equals(type)) {
            return StringResource.getNotModified();
        } else if ("badrequest".equals(type)) {
            return StringResource.getBadRequest();
        } else if ("notfound".equals(type)) {
            return StringResource.getNotFound();
        }
        return null;
    }

//    private boolean useProcessor(HttpRequestHeader requestHeader,
//            Processor processor) throws IOException {
//        processingResource = processor.onRequest(requestHeader, browser);

    // [nl] Resource̎擾Processor[vōs悤ɕύX
    private boolean useResource(HttpRequestHeader requestHeader,
            Resource resource) throws IOException {
        processingResource = resource;

        if (stopped) {
            // ̒~v`FbNprocessingResource擾
            // ɖ΂ȂȂ
            return false;
        }

        if (processingResource == null) {
            throw new HttpIOException(
                    "request processor failed to handle request:\r\n"
                            + requestHeader);
        }

        try { // ensure (processingResource == null) after the transfer.
            requestHeader.removeHopByHopHeaders();
            return processingResource.transferTo(browser, requestHeader, config);
        } finally {
            processingResource = null;
        }
    }

    private boolean isMatchToEntry(ProcessorEntry entry,
            HttpRequestHeader requestHeader) {
        if (entry.getMethod() != null) {
            if (!entry.getMethod().equals(requestHeader.getMethod())) {
                return false;
            }
        }

        if (entry.getUri() != null) {
            Matcher m = entry.getUri().matcher(requestHeader.getURI());
            if (!m.lookingAt()) {
                return false;
            }
            // TODO }b`ʂProcessorōėpł悤ɂč
        }

        return true;
    }

    private void notifyCompletion() {
        setChanged();
        notifyObservers();
    }

    public void addProcessor(String method, Pattern url, Processor processor) {
        if (processor == null) {
            throw new IllegalArgumentException("processor must not be null");
        }

        ProcessorEntry entry = new ProcessorEntry(method, url, processor);

        processorEntries.add(entry);
    }

    /**
     * stop blocking operation.
     */
    public void stop() {
        stopped = true;

        CloseUtil.close(browser);

        try {
            processingResource.stopTransfer();
        } catch (NullPointerException npe) {
            // processingResourcȅ͕ʃXbhōŝ
            // null`FbNȂĂяoAg~bNɍsȂ΂ȂȂ
            // ꂾƃbNǗGɂȂ̂optimisticɍs
            // nullꍇ͊ɓ]IĂ̂OK
            Logger.debugWithThread(npe);
        }
    }

    // [nl]
    public void addRequestFilter(RequestFilter filter) {
        if (filter == null) {
            throw new IllegalArgumentException("filter must not be null");
        }
        filterEntries.add(filter);
    }
}
