package dareka.processor.util;

import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.processor.URLResource;
import dareka.processor.URLResourceCache;
import dareka.processor.impl.Cache;
import dareka.processor.impl.NLShared;
import dareka.processor.impl.NicoApiUtil;
import dareka.processor.impl.NicoCachingTitleRetriever;

public class GetThumbInfoUtil {
    
    private static URLResourceCache getthumbinfoCache =
        new URLResourceCache("cacheGetThumbInfoMax", "cacheGetThumbInfoExpire");
    
    private static final Pattern VIDEO_ID_PATTERN = Pattern.compile(
            "<video_id>([a-z]{2}(\\d+))</video_id>");
    private static final Pattern THUMBNAIL_URL_PATTERN = Pattern.compile(
            "<thumbnail_url>.+/smile\\?i=(\\d+)</thumbnail_url>");
    private static final NLShared NLSHARED = NLShared.getInstance();
    
    /**
     * api/getthumbinfoĂяoʂXMLsmid擾B
     * @param content XML
     * @return 擾smidA݂Ȃnull
     */
    public static String getSmid(String content) {
        if (content != null) {
            Matcher m = VIDEO_ID_PATTERN.matcher(content);
            if (m.find()) {
                String smid = m.group(1), id = m.group(2);
                // LbV݂ĂTlIDƈقȂȂL^Ă
                if (Cache.getPreferredCachedVideo(smid) != null) {
                    m = THUMBNAIL_URL_PATTERN.matcher(content);
                    if (m.find() && !m.group(1).equals(id)) {
                        NLSHARED.put_vid2cid(m.group(1), id);
                    }
                }
                return smid;
            }
        }
        return null;
    }

    /**
     * api/getthumbinfoĂяoʂXML^Cg擾B
     * @param content XML
     * @return 擾^CgA݂Ȃnull
     */
    public static String getTitle(String content) {
        if (content != null) {
            return NicoCachingTitleRetriever.getTitleFromXml(content);
        }
        return null;
    }

    /**
     * api/getthumbinfoĂяoČʂXML擾B
     * ɃLbVɂ΂̌ʂԂB
     * ܂Asmid擾oꍇ̓^CgLbVXVB
     * 
     * @param id 擾ID(smid or threadId)
     * @return 擾XML(Rewriterς)Asnull
     * @throws IOException
     */
    public static String get(String id) throws IOException {
        URLResource r = getURLResource(id);
        if (r == null) {
            return null;
        }
        byte[] responseBody = r.getResponseBody();
        if (responseBody == null) {
            return null;
        }
        String content = new String(responseBody, "UTF-8");
//      String content = Main.getRewriterProcessor().stringRewriter(
//              NicoApiUtil.getThumbURL("", id),
//              new String(responseBody, "UTF-8"),
//              null, r.getResponseHeader(null, null));
// TODO RewriterĂ΂Ɩ[vɊׂ\Av
        String smid = getSmid(content);
        if (smid != null) {
            if (id.matches("\\d+")) {
                NLSHARED.putThread2Smid(id, smid);
            }
            NicoCachingTitleRetriever.putTitleCache(smid, getTitle(content));
        }
        return content;
    }

    /**
     * gC{@link #get(String) get}
     * 
     * @param id 擾ID(smid or threadId)
     * @param num gC
     * @param ms gCԊu(~b)
     * @return 擾XMLAsnull
     * @throws IOException
     */
    public static String getWithRetry(String id, int num, long ms)
            throws IOException {
        if (!isValidSmid(id) && !isValidThreadId(id)) return null;
        String content = null;
        for (int i = 0; i < num; i++) {
            content = get(id);
            if (content != null) break;
            try {
                Thread.sleep(ms);
            } catch (InterruptedException e) { }
        }
        return content;
    }

    /**
     * api/getthumbinfoĂяoĎMς
     * {@link dareka.processor.URLResource URLResource} 擾B
     * ɃLbVɂ΂̌ʂԂBȂA
     * {@link #get(String)} Ƃ͈قȂ^CgLbV͍XVȂB
     * 
     * @param id 擾ID(smid or threadId)
     * @return 擾URLResourceAMłȂnull
     */
    public static URLResource getURLResource(String id) {
        if (!isValidSmid(id) && !isValidThreadId(id)) return null;
        String url = NicoApiUtil.getThumbURL("", id);
        return getthumbinfoCache.cacheAndGet(id, url);
    }

    /**
     * XMLLbVĂ邩H
     * 
     * @param id ΏۂID(smid or threadId)
     * @return LbVĂtrue
     * @since NicoCache_nl+110706mod
     */
    public static boolean isCached(String id) {
        return getthumbinfoCache.isCached(id);
    }

    /**
     * Xapi/getthumbinfoLbVeɂČÂΔjB
     */
    public static void expires() {
        getthumbinfoCache.expires();
    }

    /**
     * api/getthumbinfoLbV폜B
     * 
     * @param id 폜api/getthumbinfoLbVID
     */
    public static void remove(String id) {
        getthumbinfoCache.remove(id);
    }

    private static final Pattern SMID_PATTERN = Pattern.compile("^\\w{2}\\d+$");
    private static final Pattern THREADID_PATTERN = Pattern.compile("^\\d+$");
    
    private static boolean isValidSmid(String id) {
        if (id == null) return false;
        Matcher m = SMID_PATTERN.matcher(id);
        return m.matches();
    }

    private static boolean isValidThreadId(String id) {
        if (id == null) return false;
        Matcher m = THREADID_PATTERN.matcher(id);
        return m.matches();
    }
}
